import pytest
from django.core.exceptions import ValidationError

from drip.models import Drip, QuerySetRule

pytestmark = pytest.mark.django_db


class TestCaseRules:
    def setup_method(self, test_method):
        self.drip = Drip.objects.create(
            name="A Drip just for Rules",
            subject_template="Hello",
            body_html_template="KETTEHS ROCK!",
        )

    def test_valid_rule(self):
        rule = QuerySetRule(
            drip=self.drip,
            field_name="date_joined",
            lookup_type="lte",
            field_value="now-60 days",
        )
        rule.clean()

    @pytest.mark.parametrize(
        "field_name, lookup_type, field_value",
        (
            ("date__joined", "lte", "now-60 days"),  # test_bad_field_name
            ("date_joined", "lte", "now-2 months"),  # test_bad_field_value
        ),
    )
    def test_raise_errors(self, field_name: str, lookup_type: str, field_value: str):
        rule = QuerySetRule(
            drip=self.drip,
            field_name=field_name,
            lookup_type=lookup_type,
            field_value=field_value,
        )
        with pytest.raises(ValidationError):
            rule.clean()



from datetime import timedelta
from typing import Any, Dict, Optional
from unittest.mock import patch

import pytest
from django.conf import settings
from django.contrib.admin.sites import AdminSite
from django.core.management import call_command
from django.test import TestCase
from django.test.client import RequestFactory
from django.urls import resolve, reverse
from django.utils import timezone

from credits.models import Profile
from drip.admin import DripAdmin
from drip.drips import DEFAULT_DRIP_MESSAGE_CLASS, DripBase, configured_message_classes
from drip.models import Drip, QuerySetRule, SentDrip
from drip.scheduler.cron_scheduler import cron_send_drips
from drip.utils import get_user_model, unicode

pytestmark = pytest.mark.django_db


def get_user_model_mock():
    from drip.models import TestUserUUIDModel

    return TestUserUUIDModel


User = get_user_model()

DEFAULT_MESSAGE_CLASSES_LENGTH = len(configured_message_classes().items())


class SetupDataDripMixin:
    NUM_STRING = [
        "first",
        "second",
        "third",
        "fourth",
        "fifth",
        "sixth",
        "seventh",
        "eighth",
        "ninth",
        "tenth",
    ]

    def build_user_data(self):
        """
        Creates 20 users, half of which buy 25 credits a day,
        and the other half that does none.
        """
        start = timezone.now() - timedelta(hours=2)
        for i, name in enumerate(self.NUM_STRING):
            user = User.objects.create(
                username="{name}_25_credits_a_day".format(name=name),
                email="{name}@test.com".format(name=name),
            )
            User.objects.filter(id=user.id).update(
                date_joined=start - timedelta(days=i),
            )

            profile = Profile.objects.get(user=user)
            profile.credits = i * 25
            profile.save()

        for i, name in enumerate(self.NUM_STRING):
            user = User.objects.create(
                username="{name}_no_credits".format(name=name),
                email="{name}@test.com".format(name=name),
            )
            User.objects.filter(id=user.id).update(
                date_joined=start - timedelta(days=i),
            )

    def build_joined_date_drip(self, shift_one=7, shift_two=8):
        model_drip = Drip.objects.create(
            name="A Custom Week Ago",
            subject_template="HELLO {{ user.username }}",
            body_html_template="KETTEHS ROCK!",
        )
        QuerySetRule.objects.create(
            drip=model_drip,
            field_name="date_joined",
            lookup_type="lt",
            field_value="now-{shift_one} days".format(shift_one=shift_one),
        )
        QuerySetRule.objects.create(
            drip=model_drip,
            field_name="date_joined",
            lookup_type="gte",
            field_value="now-{shift_two} days".format(shift_two=shift_two),
        )
        return model_drip


class TestCaseDrips(SetupDataDripMixin):
    def setup_method(self, test_method):
        self.build_user_data()

    def test_users_exists(self):
        assert 20 == User.objects.all().count()

    @pytest.mark.parametrize(
        "start_days, end_days, filter_dict, count_users",
        (
            (1, 0, {}, 2),  # test_day_zero_users
            (3, 2, {"profile__credits__gt": 0}, 1),  # test_day_two_users_active
            (3, 2, {"profile__credits": 0}, 1),  # test_day_two_users_inactive
            (8, 7, {"profile__credits__gt": 0}, 1),  # test_day_seven_users_active
            (8, 7, {"profile__credits": 0}, 1),  # test_day_seven_users_inactive
            (15, 14, {"profile__credits__gt": 0}, 0),  # test_day_fourteen_users_active
            (15, 14, {"profile__credits": 0}, 0),  # test_day_fourteen_users_inactive
        ),
    )
    def test_multiple_days_users_filter(
        self, start_days: int, end_days: int, filter_dict: Dict[str, Any], count_users: int
    ):
        start = timezone.now() - timedelta(days=start_days)
        end = timezone.now() - timedelta(days=end_days)
        assert (
            count_users
            == User.objects.filter(
                date_joined__range=(start, end),
                **filter_dict,
            ).count()
        )

    ########################
    #   RELATION SNAGGER   #
    ########################

    def test_get_simple_fields(self):
        from drip.utils import get_simple_fields

        simple_fields = get_simple_fields(User)
        assert bool([sf for sf in simple_fields if "profile" in sf[0]])

    ##################
    #   TEST DRIPS   #
    ##################

    def test_backwards_drip_class(self):
        for drip in Drip.objects.all():
            assert issubclass(drip.drip.__class__, DripBase)

    def test_custom_drip(self):
        """
        Test a simple
        """
        model_drip = self.build_joined_date_drip()
        drip = model_drip.drip

        # ensure we are starting from a blank slate
        # 2 people meet the criteria
        assert 2 == drip.get_queryset().count()
        drip.prune()
        # no one is pruned, never sent before
        assert 2 == drip.get_queryset().count()
        # confirm nothing sent before
        assert 0 == SentDrip.objects.count()

        # send the drip
        drip.send()
        assert 2 == SentDrip.objects.count()  # got sent

        for sent in SentDrip.objects.all():
            assert "HELLO" in sent.subject
            assert "KETTEHS ROCK" in sent.body

        # subsequent runs reflect previous activity
        drip = Drip.objects.get(id=model_drip.id).drip
        # 2 people meet the criteria
        assert 2 == drip.get_queryset().count()
        drip.prune()
        assert 0 == drip.get_queryset().count()  # everyone is pruned

    def test_custom_short_term_drip(self):
        model_drip = self.build_joined_date_drip(shift_one=3, shift_two=4)
        drip = model_drip.drip

        # ensure we are starting from a blank slate
        # 2 people meet the criteria
        assert 2 == drip.get_queryset().count()

    def test_custom_date_range_walk(self):
        model_drip = self.build_joined_date_drip()
        drip = model_drip.drip

        # vanilla (now-8, now-7), past (now-8-3, now-7-3),
        # future (now-8+1, now-7+1)
        for count, shifted_drip in zip([0, 2, 2, 2, 2], drip.walk(into_past=3, into_future=2)):
            assert count == shifted_drip.get_queryset().count()

        # no reason to change after a send...
        drip.send()
        drip = Drip.objects.get(id=model_drip.id).drip

        # vanilla (now-8, now-7), past (now-8-3, now-7-3),
        # future (now-8+1, now-7+1)
        for count, shifted_drip in zip([0, 2, 2, 2, 2], drip.walk(into_past=3, into_future=2)):
            assert count == shifted_drip.get_queryset().count()

    def test_custom_drip_with_count(self):
        model_drip = self.build_joined_date_drip()
        QuerySetRule.objects.create(
            drip=model_drip,
            field_name="profile__credits",
            lookup_type="gte",
            field_value="5",
        )
        drip = model_drip.drip

        # 1 person meet the criteria
        assert 1 == drip.get_queryset().count()

        for count, shifted_drip in zip([0, 1, 1, 1, 1], drip.walk(into_past=3, into_future=2)):
            assert count == shifted_drip.get_queryset().count()

    def test_exclude_and_include(self):
        model_drip = Drip.objects.create(
            name="A Custom Week Ago",
            subject_template="HELLO {{ user.username }}",
            body_html_template="KETTEHS ROCK!",
        )
        QuerySetRule.objects.create(
            drip=model_drip,
            field_name="profile__credits",
            lookup_type="gte",
            field_value="1",
        )
        QuerySetRule.objects.create(
            drip=model_drip,
            field_name="profile__credits",
            method_type="exclude",
            lookup_type="exact",
            field_value=100,
        )
        QuerySetRule.objects.create(
            drip=model_drip,
            field_name="profile__credits",
            method_type="exclude",
            lookup_type="exact",
            field_value=125,
        )
        # 7 people meet the criteria
        assert 7 == model_drip.drip.get_queryset().count()

    def test_custom_drip_static_datetime(self):
        model_drip = self.build_joined_date_drip()
        QuerySetRule.objects.create(
            drip=model_drip,
            field_name="date_joined",
            lookup_type="lte",
            field_value=(timezone.now() - timedelta(days=8)).strftime("%Y-%m-%d %H:%M:%S"),
        )
        drip = model_drip.drip

        for count, shifted_drip in zip([0, 2, 2, 0, 0], drip.walk(into_past=3, into_future=2)):
            assert count == shifted_drip.get_queryset().count()

    def test_custom_drip_static_now_datetime(self):
        model_drip = Drip.objects.create(
            name="A Custom Week Ago",
            subject_template="HELLO {{ user.username }}",
            body_html_template="KETTEHS ROCK!",
        )
        QuerySetRule.objects.create(
            drip=model_drip,
            field_name="date_joined",
            lookup_type="gte",
            field_value=(timezone.now() - timedelta(days=1)).strftime("%Y-%m-%d 00:00:00"),
        )
        drip = model_drip.drip

        # catches "today and yesterday" users
        for count, shifted_drip in zip([4, 4, 4, 4, 4], drip.walk(into_past=3, into_future=3)):
            assert count == shifted_drip.get_queryset().count()

    def test_admin_timeline_prunes_user_output(self):
        """
        multiple users in timeline is confusing.
        """
        admin = User.objects.create(username="admin", email="admin@example.com")
        admin.is_staff = True
        admin.is_superuser = True
        admin.save()

        # create a drip campaign that will surely give us duplicates.
        model_drip = Drip.objects.create(
            name="A Custom Week Ago",
            subject_template="HELLO {{ user.username }}",
            body_html_template="KETTEHS ROCK!",
        )
        QuerySetRule.objects.create(
            drip=model_drip,
            field_name="date_joined",
            lookup_type="gte",
            field_value=(timezone.now() - timedelta(days=1)).strftime("%Y-%m-%d 00:00:00"),
        )

        # then get it's admin view.
        rf = RequestFactory()
        timeline_url = reverse(
            "admin:drip_timeline",
            kwargs={
                "drip_id": model_drip.id,
                "into_past": 3,
                "into_future": 3,
            },
        )

        request = rf.get(timeline_url)
        request.user = admin

        match = resolve(timeline_url)

        response = match.func(request, *match.args, **match.kwargs)

        # check that our admin (not excluded from test) is shown once.
        assert 1 == unicode(response.content).count(admin.email)

    ##################
    #   TEST M2M     #
    ##################

    def test_annotated_field_name_property_no_count(self):
        model_drip = Drip.objects.create(
            name="A Custom Week Ago",
            subject_template="HELLO {{ user.username }}",
            body_html_template="KETTEHS ROCK!",
        )

        qsr = QuerySetRule.objects.create(
            drip=model_drip,
            field_name="date_joined",
            lookup_type="exact",
            field_value=2,
        )
        assert qsr.annotated_field_name == "date_joined"

    def test_annotated_field_name_property_with_count(self):

        model_drip = Drip.objects.create(
            name="A Custom Week Ago",
            subject_template="HELLO {{ user.username }}",
            body_html_template="KETTEHS ROCK!",
        )

        qsr = QuerySetRule.objects.create(
            drip=model_drip,
            field_name="userprofile__user__groups__count",
            lookup_type="exact",
            field_value=2,
        )
        assert qsr.annotated_field_name == "num_userprofile_user_groups"

    def test_apply_annotations_no_count(self):

        model_drip = Drip.objects.create(
            name="A Custom Week Ago",
            subject_template="HELLO {{ user.username }}",
            body_html_template="KETTEHS ROCK!",
        )

        qsr: QuerySetRule = QuerySetRule.objects.create(
            drip=model_drip,
            field_name="date_joined",
            lookup_type="exact",
            field_value=(timezone.now()).strftime("%Y-%m-%d 00:00:00"),
        )
        base_queryset = model_drip.drip.get_queryset()
        qs = qsr.apply_any_annotation(base_queryset)

        assert qs == base_queryset

    def test_apply_annotations_with_count(self):

        model_drip = Drip.objects.create(
            name="A Custom Week Ago",
            subject_template="HELLO {{ user.username }}",
            body_html_template="KETTEHS ROCK!",
        )

        qsr: QuerySetRule = QuerySetRule.objects.create(
            drip=model_drip,
            field_name="profile__user__groups__count",
            lookup_type="exact",
            field_value=2,
        )

        qs = qsr.apply_any_annotation(model_drip.drip.get_queryset())
        assert list(qs.query.annotation_select.keys()) == ["num_profile_user_groups"]  # type: ignore

    def test_apply_multiple_rules_with_aggregation(self):

        model_drip = Drip.objects.create(
            name="A Custom Week Ago",
            subject_template="HELLO {{ user.username }}",
            body_html_template="KETTEHS ROCK!",
        )

        qsr = QuerySetRule.objects.create(
            drip=model_drip,
            field_name="profile__user__groups__count",
            lookup_type="exact",
            field_value="0",
        )

        QuerySetRule.objects.create(
            drip=model_drip,
            field_name="date_joined",
            lookup_type="gte",
            field_value=(timezone.now() - timedelta(days=1)).strftime("%Y-%m-%d 00:00:00"),
        )

        qsr.clean()
        qs = model_drip.drip.apply_queryset_rules(model_drip.drip.get_queryset())
        assert qs.count() == 4

    def test_apply_and_or_queryset_ruletype(self):

        model_drip = Drip.objects.create(
            name="A Custom Week Ago",
            subject_template="HELLO {{ user.username }}",
            body_html_template="KETTEHS ROCK!",
        )

        qsr = QuerySetRule.objects.create(
            drip=model_drip,
            field_name="profile__user__groups__count",
            lookup_type="exact",
            field_value="0",
        )

        QuerySetRule.objects.create(
            drip=model_drip,
            field_name="date_joined",
            lookup_type="gte",
            field_value=(timezone.now() - timedelta(days=1)).strftime("%Y-%m-%d 00:00:00"),
            rule_type="or",
        )

        qsr.clean()
        qs = model_drip.drip.apply_queryset_rules(model_drip.drip.get_queryset())

        assert qs.count() == 20

    def test_apply_or_queryset_ruletype(self):

        model_drip = Drip.objects.create(
            name="A Custom Week Ago",
            subject_template="HELLO {{ user.username }}",
            body_html_template="KETTEHS ROCK!",
        )

        # returns 9 entries
        qsr = QuerySetRule.objects.create(
            drip=model_drip,
            field_name="profile__credits",
            lookup_type="gte",
            field_value="5",
            rule_type="or",
        )
        # returns 4 entries
        QuerySetRule.objects.create(
            drip=model_drip,
            field_name="date_joined",
            lookup_type="gte",
            field_value=(timezone.now() - timedelta(days=1)).strftime("%Y-%m-%d 00:00:00"),
            rule_type="or",
        )
        qsr.clean()
        qs = model_drip.drip.apply_queryset_rules(model_drip.drip.get_queryset())

        assert qs.count() == 12

    @pytest.mark.parametrize(
        "message_class_config, lenght_plus, default_class, custom_class",
        (
            (
                {"non-default-class": "drip.drips.OtherDripClass"},
                1,
                DEFAULT_DRIP_MESSAGE_CLASS,
                None,
            ),  # adding a brand new MessageClass
            (
                {"default": "drip.drips.OtherDripClass"},
                0,
                "drip.drips.OtherDripClass",
                None,
            ),  # Replacing an existing Message Class
            (
                {
                    "default": "drip.drips.OtherDripClass",
                    "custom": "custom.module.ClassName",
                },
                1,
                "drip.drips.OtherDripClass",
                "custom.module.ClassName",
            ),  # Mixing replacing and adding a new class
        ),
    )
    def test_message_class_for(
        self,
        message_class_config: Dict[str, str],
        lenght_plus: int,
        default_class: str,
        custom_class: Optional[str],
    ):
        setattr(
            settings,
            "DRIP_MESSAGE_CLASSES",
            message_class_config,
        )

        message_classes = configured_message_classes()

        assert len(message_classes.items()) == DEFAULT_MESSAGE_CLASSES_LENGTH + lenght_plus
        assert message_classes["default"] == default_class
        if custom_class:
            assert message_classes["custom"] == custom_class


class UrlsTestCase(TestCase):
    def test_drip_timeline_url(self):
        timeline_url = reverse(
            "admin:drip_timeline",
            kwargs={
                "drip_id": 1,
                "into_past": 2,
                "into_future": 3,
            },
        )

        assert timeline_url == "/admin/drip/drip/1/timeline/2/3/"

    def test_view_drip_email_url(self):
        view_drip_email_url = reverse(
            "admin:view_drip_email",
            kwargs={
                "drip_id": 1,
                "into_past": 2,
                "into_future": 3,
                "user_id": 4,
            },
        )

        assert view_drip_email_url == "/admin/drip/drip/1/timeline/2/3/4/"

    @patch("drip.admin.get_user_model", new=get_user_model_mock)
    def test_drip_timeline_url_user_uuid(self):
        test_admin = DripAdmin(model=Drip, admin_site=AdminSite())
        new_urls = test_admin.get_urls()
        test_url_pattern = None
        for url_pattern in new_urls:
            if url_pattern.name == "view_drip_email":
                test_url_pattern = url_pattern
                break
        assert test_url_pattern is not None
        assert (
            test_url_pattern.pattern._route  # type: ignore
            == "<int:drip_id>/timeline/<int:into_past>/<int:into_future>/<uuid:user_id>/"
        )


@pytest.fixture(scope="session")
def celery_includes():
    return [
        "drip.tasks",
    ]


class TestScheduler:
    def test_celery_beat_schedule(self, celery_app, celery_worker):
        task_name = "drip.tasks.call_send_drips_celery_command()"
        assert task_name in celery_app.conf.beat_schedule

    def test_cron_jobs_schedule(self):
        # CELERY default config
        cron_scheduler = cron_send_drips()
        assert cron_scheduler is None

        # CRON config
        setattr(
            settings,
            "DRIP_SCHEDULE_SETTINGS",
            {
                "DRIP_SCHEDULE": True,
                "DRIP_SCHEDULE_DAY_OF_WEEK": "*",
                "DRIP_SCHEDULE_HOUR": 12,
                "DRIP_SCHEDULE_MINUTE": 00,
                "SCHEDULER": "CRON",
            },
        )
        cron_scheduler = cron_send_drips()
        assert cron_scheduler
        jobs = cron_scheduler.get_jobs()
        expected_job_name = "cron_send_drips.<locals>.call_send_drips_command"
        assert expected_job_name in [job.name for job in jobs]


class TestSendDripsCommand(SetupDataDripMixin):
    @pytest.mark.parametrize(
        "build_users, model_drip_enabled, sent_drip_count, drip_count_queryset",
        (
            (True, True, 2, 2),  # Sucess case, the command send drips to users.
            (False, True, 0, 0),  # No users at all, enabled drip.
            (True, False, 0, 2),  # Disabled drip, the command will not get this drip
            (False, False, 0, 0),  # No users at all, disabled drip.
        ),
    )
    def test_send_drips_command(
        self, build_users: bool, model_drip_enabled: bool, sent_drip_count: int, drip_count_queryset: int
    ):
        if build_users:
            self.build_user_data()
        model_drip = self.build_joined_date_drip()
        model_drip.enabled = model_drip_enabled
        model_drip.save()

        call_command("send_drips")

        assert sent_drip_count == SentDrip.objects.count()
        model_drip.refresh_from_db()
        drip = model_drip.drip
        assert drip_count_queryset == drip.get_queryset().count()



import datetime
import re
from typing import Dict, Match, Union


def process_regex(matches: Match[str]) -> Dict[str, Union[str, int]]:
    matches_dict = matches.groupdict(0)
    if matches_dict["sign"] == "-":
        for k in "hours", "minutes", "seconds":
            matches_dict[k] = "-" + str(matches_dict[k])
    matches_dict.pop("sign", None)
    return matches_dict


def get_flexible_regex(string: str) -> Dict[str, Union[str, int]]:
    """
    Matches a string using a more flexible regex expression for parsing
    it to timedelta object.
    """
    matches = re.match(
        r"^((?P<weeks>-?((\d*\.\d+)|\d+))\W*w((ee)?(k(s)?)?)(,)?\W*)?"
        r"((?P<days>-?((\d*\.\d+)|\d+))\W*d(ay(s)?)?(,)?\W*)?"
        r"((?P<hours>-?((\d*\.\d+)|\d+))\W*h(ou)?(r(s)?)?(,)?\W*)?"
        r"((?P<minutes>-?((\d*\.\d+)|\d+))\W*m(in(ute)?(s)?)?(,)?\W*)?"
        r"((?P<seconds>-?((\d*\.\d+)|\d+))\W*s(ec(ond)?(s)?)?)?\W*$",
        string,
    )
    if not matches:
        raise TypeError("'{string}' is not a valid time interval".format(string=string))
    flexible_regex = matches.groupdict(0)
    return flexible_regex


def process_string(string: str) -> datetime.timedelta:
    """
    Process a string into a timedelta object, using a standard format used by
    Postgres, sqlite, or from serialization. Or a flexible regex expression
    if the standard form is not enough.
    """
    matches = re.match(
        r"^((?P<days>[-+]?\d+) days?,? )?(?P<sign>[-+]?)(?P<hours>\d+):"
        r"(?P<minutes>\d+)(:(?P<seconds>\d+(\.\d+)?))?$",
        str(string),
    )
    if matches:
        formated_dict = process_regex(matches)
    else:
        formated_dict = get_flexible_regex(string)
    return datetime.timedelta(**dict(((k, float(v)) for k, v in formated_dict.items())))


def parse(string: str) -> datetime.timedelta:
    """
    Parse a string into a timedelta object.
    """
    string = string.strip()

    if string == "":
        raise TypeError("'{string}' is not a valid time interval".format(string=string))
    return process_string(string)



# Generated by Django 3.2.14 on 2022-07-13 13:17

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('drip', '0003_testuseruuidmodel'),
    ]

    operations = [
        migrations.CreateModel(
            name='Campaign',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=256)),
                ('delete_drips', models.BooleanField(default=True)),
            ],
        ),
        migrations.AddField(
            model_name='drip',
            name='campaign',
            field=models.ForeignKey(blank=True, default=None, help_text='If set, this is the campaign to which this Drip belongs to.', null=True, on_delete=django.db.models.deletion.SET_DEFAULT, to='drip.campaign'),
        ),
    ]



import json
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

from django import forms
from django.contrib import admin
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpRequest, HttpResponse
from django.shortcuts import get_object_or_404, render
from django.urls import URLPattern, path

from drip.campaigns.admin import CampaignAdmin
from drip.drips import configured_message_classes, message_class_for
from drip.models import Campaign, Drip, QuerySetRule, SentDrip
from drip.utils import get_simple_fields, get_user_model

User = get_user_model()


class QuerySetRuleInline(admin.TabularInline):
    model = QuerySetRule


class DripForm(forms.ModelForm):
    message_class = forms.ChoiceField(
        choices=((k, "{k} ({v})".format(k=k, v=v)) for k, v in configured_message_classes().items()),
    )

    class Meta:
        model = Drip
        exclude: List[str] = []


class DripAdmin(admin.ModelAdmin):
    list_display = ("name", "enabled", "message_class")
    inlines = [
        QuerySetRuleInline,
    ]
    form = DripForm
    users_fields: Union[str, List[str]] = []

    def av(self, view: Callable) -> Callable:
        return self.admin_site.admin_view(view)

    def timeline(
        self,
        request: WSGIRequest,
        drip_id: int,
        into_past: int,
        into_future: int,
    ) -> HttpResponse:
        """
        Return a list of people who should get emails.
        """

        drip = get_object_or_404(Drip, id=drip_id)

        shifted_drips = []
        seen_users: Set[int] = set()
        for shifted_drip in drip.drip.walk(into_past=int(into_past), into_future=int(into_future) + 1):
            shifted_drip.prune()
            shifted_drips.append(
                {
                    "drip": shifted_drip,
                    "qs": shifted_drip.get_queryset().exclude(
                        id__in=seen_users,
                    ),
                },
            )
            seen_users.update(shifted_drip.get_queryset().values_list("id", flat=True))

        return render(request, "drip/timeline.html", locals())

    def get_mime_html_from_alternatives(self, alternatives: List[Tuple[str, str]]) -> Tuple[str, str]:
        html = ""
        mime = ""
        for body, mime in alternatives:
            if mime == "text/html":
                html = body
                mime = "text/html"
        return html, mime

    # Ignoring this line because mypy says User is not a valid type
    def get_mime_html(self, drip: Drip, user: User) -> Tuple[str, str]:  # type: ignore
        # Ignoring this line because mypy says DripMessage is not callable
        drip_message = message_class_for(  # type: ignore
            drip.message_class,
        )(drip.drip, user)
        if drip_message.message.alternatives:
            return self.get_mime_html_from_alternatives(drip_message.message.alternatives)
        html = drip_message.message.body
        mime = "text/plain"
        return html, mime

    def view_drip_email(
        self,
        request: WSGIRequest,
        drip_id: int,
        into_past: int,
        into_future: int,
        user_id: int,
    ) -> HttpResponse:
        drip = get_object_or_404(Drip, id=drip_id)
        user = get_object_or_404(User, id=user_id)

        html, mime = self.get_mime_html(drip, user)

        return HttpResponse(html, content_type=mime)

    def build_extra_context(self, extra_context: Optional[Dict[str, Any]]) -> Dict[str, Any]:
        extra_context = extra_context or {}
        User = get_user_model()
        if not self.users_fields:
            self.users_fields = json.dumps(get_simple_fields(User))
        extra_context["field_data"] = self.users_fields
        return extra_context

    def add_view(self, request: HttpRequest, form_url: str = "", extra_context: Any = None) -> HttpResponse:
        return super(DripAdmin, self).add_view(
            request,
            form_url=form_url,
            extra_context=self.build_extra_context(extra_context),  # type: ignore
        )

    def change_view(self, request: HttpRequest, object_id: str, form_url="", extra_context: Any = None) -> HttpResponse:
        return super(DripAdmin, self).change_view(
            request, object_id, form_url=form_url, extra_context=self.build_extra_context(extra_context)
        )

    def get_urls(self) -> List[URLPattern]:
        urls = super(DripAdmin, self).get_urls()
        my_urls = [
            path(
                "<int:drip_id>/timeline/<int:into_past>/<int:into_future>/",
                self.av(self.timeline),
                name="drip_timeline",
            ),
        ]

        User = get_user_model()
        if User._meta.get_field("id").get_internal_type() == "UUIDField":
            my_urls += [
                path(
                    "<int:drip_id>/timeline/<int:into_past>/<int:into_future>/<uuid:user_id>/",  # noqa
                    self.av(self.view_drip_email),
                    name="view_drip_email",
                ),
            ]
        else:
            my_urls += [
                path(
                    "<int:drip_id>/timeline/<int:into_past>/<int:into_future>/<int:user_id>/",  # noqa
                    self.av(self.view_drip_email),
                    name="view_drip_email",
                ),
            ]

        return my_urls + urls


admin.site.register(Drip, DripAdmin)


class SentDripAdmin(admin.ModelAdmin):
    list_display = [f.name for f in SentDrip._meta.fields]
    ordering = ["-id"]


admin.site.register(SentDrip, SentDripAdmin)


admin.site.register(Campaign, CampaignAdmin)



import random
from typing import Any, List

import pytest

from drip.models import Campaign, Drip
from drip.tests.factories import CampaignFactory, DripFactory

pytestmark = pytest.mark.django_db


def _drips_generator(amount: int, **drip_extra_args: Any) -> List[Campaign]:
    return DripFactory.create_batch(amount, **drip_extra_args)


class TestCaseCampaign:
    def test_campaings_creation(self):
        campaign = CampaignFactory()
        drip_count = random.randint(5, 15)
        _drips_generator(drip_count, campaign=campaign)

        assert len(campaign.drip_set.all()) == drip_count

    @pytest.mark.parametrize(
        ("delete_drips, drip_count, drip_count_after_delete"),
        (
            (True, 10, 0),
            (False, 10, 10),
        ),
    )
    def test_remove_campaings(
        self,
        delete_drips: bool,
        drip_count: int,
        drip_count_after_delete: int,
    ):
        campaign = CampaignFactory(delete_drips=delete_drips)
        _drips_generator(drip_count, campaign=campaign)

        campaign.delete()

        assert Drip.objects.count() == drip_count_after_delete



from django.conf import settings

SCHEDULER_CRON = "CRON"
SCHEDULER_CELERY = "CELERY"
VALID_SCHEDULERS = (
    SCHEDULER_CRON,
    SCHEDULER_CELERY,
)


def get_drip_scheduler_settings():
    DRIP_SCHEDULE_SETTINGS = getattr(settings, "DRIP_SCHEDULE_SETTINGS", {})
    DRIP_SCHEDULE = DRIP_SCHEDULE_SETTINGS.get("DRIP_SCHEDULE", False)
    DRIP_SCHEDULE_DAY_OF_WEEK = DRIP_SCHEDULE_SETTINGS.get("DRIP_SCHEDULE_DAY_OF_WEEK", 0)
    DRIP_SCHEDULE_HOUR = DRIP_SCHEDULE_SETTINGS.get("DRIP_SCHEDULE_HOUR", 0)
    DRIP_SCHEDULE_MINUTE = DRIP_SCHEDULE_SETTINGS.get("DRIP_SCHEDULE_MINUTE", 0)
    SCHEDULER = DRIP_SCHEDULE_SETTINGS.get("SCHEDULER", SCHEDULER_CRON)
    return DRIP_SCHEDULE, DRIP_SCHEDULE_DAY_OF_WEEK, DRIP_SCHEDULE_HOUR, DRIP_SCHEDULE_MINUTE, SCHEDULER



from celery import current_app
from celery.schedules import crontab
from django.core.management import call_command

from drip.scheduler.constants import SCHEDULER_CELERY, get_drip_scheduler_settings

(
    DRIP_SCHEDULE,
    DRIP_SCHEDULE_DAY_OF_WEEK,
    DRIP_SCHEDULE_HOUR,
    DRIP_SCHEDULE_MINUTE,
    SCHEDULER,
) = get_drip_scheduler_settings()

app = current_app._get_current_object()

CELERY_ENABLED = SCHEDULER == SCHEDULER_CELERY

if DRIP_SCHEDULE and CELERY_ENABLED:

    @app.task
    def call_send_drips_celery_command():
        call_command("send_drips")


@app.on_after_finalize.connect
def app_ready_drip(sender, **kwargs):
    if DRIP_SCHEDULE and CELERY_ENABLED:
        sender.add_periodic_task(
            crontab(
                day_of_week=DRIP_SCHEDULE_DAY_OF_WEEK,
                hour=DRIP_SCHEDULE_HOUR,
                minute=DRIP_SCHEDULE_MINUTE,
            ),
            call_send_drips_celery_command,
        )



# Generated by Django 3.0.6 on 2020-05-29 20:38

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Drip",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("date", models.DateTimeField(auto_now_add=True)),
                ("lastchanged", models.DateTimeField(auto_now=True)),
                (
                    "name",
                    models.CharField(
                        help_text="A unique name for this drip.",
                        max_length=255,
                        unique=True,
                        verbose_name="Drip Name",
                    ),
                ),
                ("enabled", models.BooleanField(default=False)),
                (
                    "from_email",
                    models.EmailField(
                        blank=True,
                        help_text="Set a custom from email.",
                        max_length=254,
                        null=True,
                    ),
                ),
                (
                    "from_email_name",
                    models.CharField(
                        blank=True,
                        help_text="Set a name for a custom from email.",
                        max_length=150,
                        null=True,
                    ),
                ),
                ("subject_template", models.TextField(blank=True, null=True)),
                (
                    "body_html_template",
                    models.TextField(
                        blank=True,
                        help_text="You will have settings and user in the context.",
                        null=True,
                    ),
                ),
                (
                    "message_class",
                    models.CharField(blank=True, default="default", max_length=120),
                ),
            ],
        ),
        migrations.CreateModel(
            name="SentDrip",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("date", models.DateTimeField(auto_now_add=True)),
                ("subject", models.TextField()),
                ("body", models.TextField()),
                (
                    "from_email",
                    models.EmailField(default=None, max_length=254, null=True),
                ),
                (
                    "from_email_name",
                    models.CharField(default=None, max_length=150, null=True),
                ),
                (
                    "drip",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="sent_drips",
                        to="drip.Drip",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="sent_drips",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="QuerySetRule",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("date", models.DateTimeField(auto_now_add=True)),
                ("lastchanged", models.DateTimeField(auto_now=True)),
                (
                    "method_type",
                    models.CharField(
                        choices=[("filter", "Filter"), ("exclude", "Exclude")],
                        default="filter",
                        max_length=12,
                    ),
                ),
                (
                    "field_name",
                    models.CharField(max_length=128, verbose_name="Field name of User"),
                ),
                (
                    "lookup_type",
                    models.CharField(
                        choices=[
                            ("exact", "exactly"),
                            ("iexact", "exactly (case insensitive)"),
                            ("contains", "contains"),
                            ("icontains", "contains (case insensitive)"),
                            ("regex", "regex"),
                            ("iregex", "contains (case insensitive)"),
                            ("gt", "greater than"),
                            ("gte", "greater than or equal to"),
                            ("lt", "less than"),
                            ("lte", "less than or equal to"),
                            ("startswith", "starts with"),
                            ("endswith", "starts with"),
                            ("istartswith", "ends with (case insensitive)"),
                            ("iendswith", "ends with (case insensitive)"),
                        ],
                        default="exact",
                        max_length=12,
                    ),
                ),
                (
                    "field_value",
                    models.CharField(
                        help_text="Can be anything from a number, to a string. Or, do `now-7 days` or `today+3 days` for fancy timedelta.",
                        max_length=255,
                    ),
                ),
                (
                    "drip",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="queryset_rules",
                        to="drip.Drip",
                    ),
                ),
            ],
        ),
    ]



from apscheduler.schedulers.background import BackgroundScheduler
from django.conf import settings
from django.core.management import call_command

from drip.scheduler.constants import SCHEDULER_CRON, get_drip_scheduler_settings


def cron_send_drips():
    def call_send_drips_command():
        call_command("send_drips")

    (
        DRIP_SCHEDULE,
        DRIP_SCHEDULE_DAY_OF_WEEK,
        DRIP_SCHEDULE_HOUR,
        DRIP_SCHEDULE_MINUTE,
        SCHEDULER,
    ) = get_drip_scheduler_settings()
    CRON_ENABLED = SCHEDULER == SCHEDULER_CRON
    if DRIP_SCHEDULE and CRON_ENABLED:
        cron_scheduler = BackgroundScheduler()
        cron_scheduler.add_job(
            call_send_drips_command,
            "cron",
            day_of_week=DRIP_SCHEDULE_DAY_OF_WEEK,
            hour=DRIP_SCHEDULE_HOUR,
            minute=DRIP_SCHEDULE_MINUTE,
            timezone=getattr(settings, "TIME_ZONE", "UTC"),
        )
        cron_scheduler.start()
        return cron_scheduler



from typing import Any, Dict, Optional, Type

import pytest
from django.conf import settings
from django.core import mail

from drip.drips import DripMessage
from drip.models import Drip, QuerySetRule
from drip.utils import get_user_model

pytestmark = pytest.mark.django_db
User = get_user_model()


# Used by CustomMessagesTest
class PlainDripEmail(DripMessage):
    @property
    def message(self):
        if not self._message:
            email = mail.EmailMessage(
                self.subject,
                self.plain,
                self.from_email,
                [self.user.email],
            )
            self._message = email
        return self._message


class TestCustomMessages:
    old_msg_classes: Optional[str]

    @classmethod
    def setup_class(cls):
        cls.old_msg_classes = getattr(settings, "DRIP_MESSAGE_CLASSES", None)

    def setup_method(self, test_method):
        self.user = User.objects.create(
            username="customuser",
            email="custom@example.com",
        )
        self.model_drip = Drip.objects.create(
            name="A Custom Week Ago",
            subject_template="HELLO {{ user.username }}",
            body_html_template="<h2>This</h2> is an <b>example</b>" " html <strong>body</strong>.",
        )
        QuerySetRule.objects.create(
            drip=self.model_drip,
            field_name="id",
            lookup_type="exact",
            field_value=self.user.id,
        )

    @pytest.mark.parametrize(
        "settings_message_config, message_class, expected_result, len_mail, mail_instance",
        (
            (None, None, 1, 1, mail.EmailMultiAlternatives),  # test_default_email
            (
                {"plain": "drip.tests.PlainDripEmail"},
                None,
                1,
                1,
                mail.EmailMultiAlternatives,  # Since we did not specify custom class, default should be used.
            ),  # test_custom_added_not_used
            (
                {
                    "plain": "drip.tests.test_custom_messages.PlainDripEmail",
                },
                "plain",
                1,
                1,
                mail.EmailMessage,  # In this case we did specify the custom key, so message should be of custom type.
            ),  # test_custom_added_and_used
            (
                {
                    "default": "drip.tests.test_custom_messages.PlainDripEmail",
                },
                None,
                1,
                1,
                mail.EmailMessage,
            ),  # test_override_default
        ),
    )
    def test_emails(
        self,
        settings_message_config: Optional[Dict[str, Any]],
        message_class: Optional[str],
        expected_result: int,
        len_mail: int,
        mail_instance: Type[mail.EmailMessage],
    ):
        if settings_message_config:
            settings.DRIP_MESSAGE_CLASSES = settings_message_config
        if message_class:
            self.model_drip.message_class = message_class
            self.model_drip.save()
        result = self.model_drip.drip.send()

        assert expected_result == result
        assert len_mail == len(mail.outbox)

        email = mail.outbox.pop()
        assert isinstance(email, mail_instance)



from collections import OrderedDict
from typing import Callable, List, Set, Union

from django.contrib import admin
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.shortcuts import get_object_or_404, render
from django.urls import URLPattern, path

from drip.models import Campaign, Drip
from drip.utils import get_user_model

User = get_user_model()


class DripInline(admin.TabularInline):
    model = Drip


class CampaignAdmin(admin.ModelAdmin):
    inlines = [
        DripInline,
    ]
    users_fields: Union[str, List[str]] = []

    def av(self, view: Callable) -> Callable:
        return self.admin_site.admin_view(view)

    def timeline(
        self,
        request: WSGIRequest,
        drip_id: int,
        into_past: int,
        into_future: int,
    ) -> HttpResponse:
        """
        Return a list of people who should get emails.
        """

        campaign = get_object_or_404(Campaign, id=drip_id)
        new_shifted_drips = OrderedDict()
        for shift in range(-into_past, into_future + 1):
            new_shifted_drips[shift] = {"drips": [], "now_shift_kwargs_days": shift}
        for drip in campaign.drip_set.all():
            seen_users: Set[int] = set()
            for shifted_drip in drip.drip.walk(into_past=int(into_past), into_future=int(into_future) + 1):
                shifted_drip.prune()
                shifted_data = {
                    "drip_model": drip,
                    "drip": shifted_drip,
                    "qs": shifted_drip.get_queryset().exclude(
                        id__in=seen_users,
                    ),
                }
                seen_users.update(shifted_drip.get_queryset().values_list("id", flat=True))
                shift_days = shifted_drip.now_shift_kwargs.get("days")
                if shift_days:
                    new_shifted_drips[shift_days]["drips"].append(shifted_data)  # type: ignore
                    new_shifted_drips[shift_days]["now"] = shifted_drip.now()

        return render(request, "campaign/timeline.html", locals())

    def get_urls(self) -> List[URLPattern]:
        urls = super(CampaignAdmin, self).get_urls()
        my_urls = [
            path(
                "<int:drip_id>/timeline/<int:into_past>/<int:into_future>/",
                self.av(self.timeline),
                name="campaign_drip_timeline",
            ),
        ]

        return my_urls + urls



from django.db import models


class Campaign(models.Model):
    name = models.CharField(max_length=256)
    delete_drips = models.BooleanField(default=True)

    def delete(self, using=None, keep_parents=False):
        if self.delete_drips:
            self.drip_set.all().delete()
        super().delete(using, keep_parents)



from factory import Faker, SubFactory
from factory.django import DjangoModelFactory

from drip.models import Campaign, Drip


class CampaignFactory(DjangoModelFactory):
    name = Faker("word")
    delete_drips = Faker("pybool")

    class Meta:
        model = Campaign


class DripFactory(DjangoModelFactory):
    name = Faker("sentences", nb=3)
    campaign = SubFactory(CampaignFactory)

    class Meta:
        model = Drip
        django_get_or_create = ("name",)



"""Defines the types to be used within
the type annotations all along the system"""
from datetime import timedelta
from typing import TypeVar, Union

from django.db.models import F, Field, ForeignKey, ManyToManyField, OneToOneField
from django.db.models.fields.related import ForeignObjectRel as RelatedObject

AbstractQuerySetRule = TypeVar("AbstractQuerySetRule")
TimeDeltaOrStr = Union[timedelta, str]
BoolOrStr = Union[bool, str]
FExpressionOrStr = Union[F, str]
FieldValue = Union[TimeDeltaOrStr, BoolOrStr, FExpressionOrStr]
FieldType = Union[Field, ForeignKey, OneToOneField, RelatedObject, ManyToManyField]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import re
import sys
from os import path

from setuptools import setup

current_dir = path.abspath(path.dirname("__file__"))
with open(path.join(current_dir, "README.md"), encoding="utf-8") as file:
    long_description = file.read()


name = "django-drip-campaigns"
package = "drip"
description = "Drip campaigns for Django."
url = "https://github.com/rootstrap/django-drip-campaigns"
author = "Kalil de Lima"
author_email = "kalil@rootstrap.com"
license = "MIT"
install_requires = ["Django>=2.2", "apscheduler"]
keywords = "django drip email user query"


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, "__init__.py")).read()
    return re.search("^__version__ = ['\"]([^'\"]+)['\"]", init_py, re.MULTILINE).group(1)  # type: ignore


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [
        dirpath
        for dirpath, dirnames, filenames in os.walk(package)
        if os.path.exists(os.path.join(dirpath, "__init__.py"))
    ]


def get_package_data(package):
    """
    Return all files under the root package, that are not in a
    package themselves.
    """
    walk = [
        (dirpath.replace(package + os.sep, "", 1), filenames)
        for dirpath, dirnames, filenames in os.walk(package)
        if not os.path.exists(os.path.join(dirpath, "__init__.py"))
    ]

    filepaths = []
    for base, filenames in walk:
        filepaths.extend([os.path.join(base, filename) for filename in filenames])
    return {package: filepaths}


if sys.argv[-1] == "publish":
    os.system("python setup.py sdist upload")
    args = {"version": get_version(package)}
    print("You probably want to also tag the version now:")
    print("  git tag -a v{v} -m 'version v{v}'".format(v=args["version"]))
    print("  git push --tags")
    sys.exit()


setup(
    name=name,
    version=get_version(package),
    url=url,
    license=license,
    description=description,
    author=author,
    author_email=author_email,
    keywords=keywords,
    packages=get_packages(package),
    package_data=get_package_data(package),
    install_requires=install_requires,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



# Generated by Django 3.0.7 on 2020-11-25 13:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("drip", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="querysetrule",
            name="rule_type",
            field=models.CharField(
                choices=[("or", "Or"), ("and", "And")], default="and", max_length=3
            ),
        ),
    ]



from datetime import datetime, timedelta
from typing import Any, Callable, Dict, List, Type

import six
from django.contrib.auth.models import User
from django.db import models
from django.db.models import ForeignKey, ManyToManyField, OneToOneField
from django.db.models.fields.related import ForeignObjectRel as RelatedObject

from drip.scheduler.constants import VALID_SCHEDULERS, get_drip_scheduler_settings
from drip.types import FieldType

basestring = (str, bytes)
unicode = str


def check_redundant(model_stack: List[Type[models.Model]], stack_limit: int) -> bool:
    """
    Checks to ensure recursion isnt being redundant

    :param model_stack: List of models.Model
    :type model_stack: List[Type[models.Model]]
    :param stack_limit: recursion depth to check redundancy
    :type stack_limit: int
    :return: Bool that controls if stops recursion
    :rtype: bool
    """
    stop_recursion = False
    if len(model_stack) > stack_limit:
        # rudimentary CustomUser->User->CustomUser->User detection, or
        # stack depth shouldn't exceed x, or
        # we've hit a point where we are repeating models
        if (
            (model_stack[-3] == model_stack[-1])
            or (len(model_stack) > 5)
            or (len(set(model_stack)) != len(model_stack))
        ):
            stop_recursion = True
    return stop_recursion


def get_field_name(field, RelatedObject):
    field_name = field.name

    if isinstance(field, RelatedObject):
        field_name = field.field.related_query_name()
    return field_name


def get_full_field(parent_field: str, field_name: str) -> str:
    """
    :param parent_field: Name of parent field
    :type parent_field: str
    :param field_name: Field name
    :type field_name: str
    :return: It is the join between parent_field and field_name or just the field_name
    :rtype: str
    """
    if parent_field:
        full_field = "__".join([parent_field, field_name])
    else:
        full_field = field_name
    return full_field


def get_rel_model(field: FieldType, RelatedObject: RelatedObject) -> Type[models.Model]:
    if isinstance(field, RelatedObject):  # type: ignore
        RelModel = field.model
        # field_names.extend(get_fields(RelModel, full_field, True))
    else:
        RelModel = field.related_model  # type: ignore
    return RelModel  # type: ignore


def is_valid_instance(field: FieldType) -> bool:
    return (
        isinstance(field, ForeignKey)
        or isinstance(field, OneToOneField)
        or isinstance(field, RelatedObject)
        or isinstance(field, ManyToManyField)
    )


def get_out_fields(
    Model: Type[models.Model],
    parent_field: str,
    model_stack: List[Type[models.Model]],
    excludes: List[str],
    fields: List[FieldType],
):
    out_fields = []
    for field in fields:

        field_name = get_field_name(field, RelatedObject)

        full_field = get_full_field(parent_field, field_name)

        if len([True for exclude in excludes if (exclude in full_field)]):
            continue

        # add to the list
        out_fields.append([full_field, field_name, Model, field.__class__])

        if is_valid_instance(field):

            RelModel = get_rel_model(field, RelatedObject)  # type: ignore

            out_fields.extend(
                get_fields(RelModel, full_field, list(model_stack)),
            )

    return out_fields


def get_fields(
    Model: Type[models.Model],
    parent_field: str = "",
    model_stack: List[Type[models.Model]] = [],
    stack_limit: int = 2,
    excludes: List[str] = ["permissions", "comment", "content_type"],
):
    """
    Given a Model, return a list of lists of strings with important stuff:
    ...
    ['test_user__user__customuser', 'customuser', 'User', 'RelatedObject']
    ['test_user__unique_id', 'unique_id', 'TestUser', 'CharField']
    ['test_user__confirmed', 'confirmed', 'TestUser', 'BooleanField']
    ...

    """

    # github.com/omab/python-social-auth/commit/d8637cec02422374e4102231488481170dc51057
    if isinstance(Model, six.string_types):
        app_label, model_name = Model.split(".")
        Model = models.get_model(app_label, model_name)  # type: ignore

    fields = Model._meta.fields + Model._meta.many_to_many + Model._meta.get_fields()  # type: ignore
    model_stack.append(Model)

    # do a variety of checks to ensure recursion isnt being redundant

    stop_recursion = check_redundant(model_stack, stack_limit)

    if stop_recursion:
        return []  # give empty list for "extend"

    return get_out_fields(Model, parent_field, model_stack, excludes, fields)


def give_model_field(full_field: str, Model: Type[models.Model]) -> tuple:
    """Given a field_name and Model:

    "test_user__unique_id", <AchievedGoal>

    Returns "test_user__unique_id", "id", <Model>, <ModelField>

    :param full_field: full field name
    :type full_field: str
    :param Model: models.Model
    :type Model: models.Model
    :raises Exception: If the key is not found it raises and exception
    :return: It is a tuple with field full name, field name, <Model>, <ModelField>
    :rtype: tuple
    """

    field_data = get_fields(Model, "", [])

    for full_key, name, _Model, _ModelField in field_data:
        if full_key == full_field:
            return full_key, name, _Model, _ModelField
    message_exception = "Field key `{field}` not found on `{model}`.".format(
        field=full_field,
        model=Model.__name__,  # type: ignore
    )
    raise Exception(message_exception)


def get_simple_fields(Model: Type[models.Model], **kwargs) -> List:
    ret_list: List = []
    for f in get_fields(Model, **kwargs):
        if "__" in f[0]:
            # Add __user__ to the fields in related models
            parts = f[0].split("__")
            f[0] = parts[0] + "__user__" + parts[1]
        if f[0] not in [x[0] for x in ret_list]:
            # Add field if not already in list
            ret_list.append([f[0], f[3].__name__])
    return ret_list


def get_user_model() -> Type[User]:
    # handle 1.7 and back
    try:
        from django.contrib.auth import get_user_model as django_get_user_model

        User = django_get_user_model()
    except ImportError:
        pass
    return User


def get_conditional_now() -> Callable:
    # handle now import for appropiate django version
    try:
        from django.utils.timezone import now as conditional_now
    except ImportError:
        conditional_now = datetime.now
    return conditional_now


def build_now_from_timedelta(now_shift_kwargs: Dict[str, Any]) -> datetime:
    """
    Build "now" from shift timedelta.
    """
    conditional_now = get_conditional_now()
    return conditional_now() + timedelta(**now_shift_kwargs)


class DripScheduleSettingsError(Exception):
    pass


def validate_schedules():
    _, _, _, _, SCHEDULER = get_drip_scheduler_settings()
    if SCHEDULER not in VALID_SCHEDULERS:
        raise DripScheduleSettingsError(f"{SCHEDULER} is not a valid SCHEDULER configuration")



import functools
import logging
import operator
from collections import ChainMap
from datetime import datetime, timedelta
from functools import lru_cache
from importlib import import_module
from typing import Any, Dict, List, Optional, TypedDict, Union

from django.conf import settings
from django.core.mail import EmailMessage, EmailMultiAlternatives
from django.db.models import Q
from django.db.models.manager import BaseManager
from django.db.models.query import QuerySet
from django.template import Context, Template
from django.utils.html import strip_tags
from django.utils.safestring import SafeString
from typing_extensions import TypeAlias

from drip.exceptions import MessageClassNotFound
from drip.models import Drip, SentDrip
from drip.utils import build_now_from_timedelta, get_conditional_now, get_user_model

User = get_user_model()

conditional_now = get_conditional_now()


DEFAULT_DRIP_MESSAGE_CLASS = "drip.drips.DripMessage"


def configured_message_classes() -> ChainMap:
    """
    Returns a ChainMap (basically a dict), between default settings and
    settings.DRIP_MESSAGE_CLASSES which should be a dictionary of the form:

        { message_class_name: class_path }

    For example, if you define this at DRIP_MESSAGE_CLASSES:

        {
            'default': 'my.path.to.DripMessage',
            'my_other_class': 'my.other.path.to.DripMessage',
        }

    The returning value of this method would be:

        {
            'default': 'my.path.to.Drip',
            'my_other_class': 'my.other.path.to.DripMessage',
        }

    But, if you define this:

        {
            'my_other_class': 'my.other.path.to.DripMessage',
        }

    The returning value of this method would be:

        {
            'default': 'drip.drips.DripMessage',
            'my_other_class': 'my.other.path.to.DripMessage',
        }

    A 'default' key is added when it's not present on DRIP_MESSAGE_CLASSES.
    """
    default_config = {"default": DEFAULT_DRIP_MESSAGE_CLASS}
    user_defined_config = getattr(settings, "DRIP_MESSAGE_CLASSES", {})
    return ChainMap(user_defined_config, default_config)


@lru_cache()
def message_class_for(name: str) -> "DripMessage":
    """
    Given a class's path, returns a reference to it.
    Raises MessageClassNotFound exception if name is not present in
    the ChainMap produced by configured_message_classes.
    """
    try:
        path = configured_message_classes()[name]
    except KeyError:
        logging.error('Name "{}" not found in configured message classes.'.format(name))
        raise MessageClassNotFound() from None
    else:
        mod_name, klass_name = path.rsplit(".", 1)
        mod = import_module(mod_name)
        klass = getattr(mod, klass_name)
        return klass


class DripBaseParamsOptions(TypedDict):
    drip_model: "Drip"
    name: str
    now_shift_kwargs: Dict[str, Any]


class DripMessage(object):
    """
    Email message abstraction for manage message interactions, based on EmailMultiAlternatives.
    You can extend this manually overriding any method that you need.

    :param drip_base: DripBase object to build email
    :type drip_base: DripBase
    :param user: User to send email
    :type user: User
    """

    # Ignoring this line because mypy says User is not a valid type
    def __init__(self, drip_base: "DripBase", user: User):  # type: ignore
        self.drip_base = drip_base
        self.user: TypeAlias = user
        self._context: Optional[Context] = None
        self._subject: Optional[SafeString] = None
        self._body: Optional[SafeString] = None
        self._plain: str = ""
        self._message: Union[EmailMessage, EmailMultiAlternatives, None] = None

    @property
    def from_email(self) -> Optional[str]:
        return self.drip_base.from_email

    @property
    def from_email_name(self) -> Optional[str]:
        return self.drip_base.from_email_name

    @property
    def context(self) -> Context:
        if not self._context:
            self._context = Context({"user": self.user})
        return self._context

    @property
    def subject(self) -> SafeString:
        drip_subject: SafeString
        if not self._subject:
            drip_subject = Template(
                self.drip_base.subject_template,
            ).render(self.context)
            self._subject = drip_subject
        else:
            drip_subject = self._subject
        return drip_subject

    @property
    def body(self) -> SafeString:
        drip_body: SafeString
        if not self._body:
            drip_body = Template(
                self.drip_base.body_template,
            ).render(self.context)
            self._body = drip_body
        else:
            drip_body = self._body
        return drip_body

    @property
    def plain(self) -> str:
        if not self._plain:
            self._plain = strip_tags(self.body)
        return self._plain

    def get_from_(self) -> str:
        if self.drip_base.from_email_name:
            from_ = "{name} <{email}>".format(
                name=self.drip_base.from_email_name,
                email=self.drip_base.from_email,
            )
        else:
            from_ = self.drip_base.from_email
        return from_

    @property
    def message(self) -> Union[EmailMessage, EmailMultiAlternatives]:
        if not self._message:
            from_ = self.get_from_()

            self._message = EmailMultiAlternatives(
                self.subject,
                self.plain,
                from_,
                [self.user.email],
            )

            # check if there are html tags in the rendered template
            if len(self.plain) != len(self.body):
                self._message.attach_alternative(self.body, "text/html")
        return self._message


class DripBase(object):
    """
    A base object for defining a Drip.

    You can extend this manually, or you can create full querysets
    and templates from the admin.
    """

    #: needs a unique name
    name: str
    subject_template: str
    body_template: str
    from_email: str
    from_email_name: str

    def __init__(self, drip_model: Drip, *args, **kwargs):
        self.drip_model = drip_model

        self.name = kwargs.pop("name", None)
        self.from_email = kwargs.pop("from_email", None)
        self.from_email_name = kwargs.pop(
            "from_email_name",
            None,
        )
        self.subject_template = kwargs.pop(
            "subject_template",
            None,
        )
        self.body_template = kwargs.pop("body_template", None)

        if not self.name:
            raise AttributeError("You must define a name.")

        self.now_shift_kwargs = kwargs.get("now_shift_kwargs", {})

    #########################
    #   DATE MANIPULATION   #
    #########################

    def now(self) -> datetime:
        """
        This allows us to override what we consider "now", making it easy
        to build timelines of who gets what when.
        """
        return build_now_from_timedelta(self.now_shift_kwargs)

    def timedelta(self, *args, **kwargs) -> timedelta:
        """
        If needed, this allows us the ability
        to manipulate the slicing of time.
        """
        return timedelta(*args, **kwargs)

    def walk(self, into_past: int = 0, into_future: int = 0) -> List["DripBase"]:
        """Walk over a date range and create
            new instances of self with new ranges.

        :param into_past: defaults to 0
        :type into_past: int, optional
        :param into_future: defaults to 0
        :type into_future: int, optional
        :return: List of DripBase instances.
        :rtype: List[DripBase]
        """
        walked_range = []
        for shift in range(-into_past, into_future):
            kwargs: DripBaseParamsOptions = dict(
                drip_model=self.drip_model,
                name=self.name,
                now_shift_kwargs={"days": shift},
            )
            walked_range.append(self.__class__(**kwargs))
        return walked_range

    def apply_queryset_rules(self, manager_qs: Union[BaseManager, QuerySet]) -> QuerySet:
        """First collect all filter/exclude kwargs and apply any annotations.
        Then apply all filters at once, and all excludes at once.

        :param manager_qs: Base queryset or manager to apply queryset rules
        :type manager_qs: Union[BaseManager, QuerySet]
        :return: Queryset with all (AND/OR) filters applied
        :rtype: QuerySet
        """
        return self.apply_and_queryset_rules(manager_qs) | self.apply_or_queryset_rules(manager_qs)

    def apply_or_queryset_rules(self, manager_qs: Union[BaseManager, QuerySet]) -> QuerySet:
        """First collect all filter kwargs. Then apply OR filters at once.

        :param manager_qs: Base queryset or manager to apply queryset rules
        :type manager_qs: Union[BaseManager, QuerySet]
        :return: Queryset with OR filters applied
        :rtype: QuerySet
        """
        rules: List[Q] = []
        rule_set = self.drip_model.queryset_rules.filter(rule_type="or")
        for query_rule in rule_set:
            kwargs = query_rule.filter_kwargs(now=self.now)
            query_or = Q(**kwargs)
            rules.append(query_or)

        query: Q = rules.pop() if rules else Q()
        for rule in rules:
            query |= rule
        qs = manager_qs.filter(query) if query != Q() else manager_qs.none()
        return qs

    def apply_and_queryset_rules(self, manager_qs: Union[BaseManager, QuerySet]) -> QuerySet:
        """First collect all filter/exclude kwargs and apply any annotations.
        Then apply AND filters at once, and all excludes at once.

        :param manager_qs: Base queryset or manager to apply queryset rules
        :type manager_qs: Union[BaseManager, QuerySet]
        :return: Queryset with AND filters applied
        :rtype: QuerySet
        """
        clauses: Dict[str, List] = {
            "filter": [],
            "exclude": [],
        }
        rules = []
        qs = manager_qs.all()
        for rule in self.drip_model.queryset_rules.filter(rule_type="and"):
            rules.append(rule)

            clause = clauses.get(rule.method_type, clauses["filter"])

            kwargs = rule.filter_kwargs(now=self.now)
            clause.append(Q(**kwargs))

            qs = rule.apply_any_annotation(qs)

        if clauses["exclude"]:
            qs = qs.exclude(functools.reduce(operator.or_, clauses["exclude"]))

        qs = qs.filter(*clauses["filter"]) if len(rules) > 0 else qs.none()

        return qs

    ##################
    #   MANAGEMENT   #
    ##################

    def get_queryset(self) -> QuerySet:
        """Apply queryset rules or returns the existing queryset

        :return: Queryset with all (AND/OR) filters applied
        :rtype: QuerySet
        """
        queryset = getattr(self, "_queryset", None)
        if queryset is None:
            self._queryset = self.apply_queryset_rules(
                self.queryset(),
            ).distinct()
        return self._queryset

    def run(self) -> Optional[int]:
        """Get the queryset, prune sent people, and send it.

        :return: Returns count of created SentDrips.
        :rtype: Optional[int]
        """
        if not self.drip_model.enabled:
            return None

        self.prune()
        count = self.send()

        return count

    def prune(self) -> None:
        """Do an exclude for all Users who have a SentDrip already."""
        target_user_ids = self.get_queryset().values_list("id", flat=True)
        exclude_user_ids = SentDrip.objects.filter(
            date__lt=conditional_now(),
            drip=self.drip_model,
            user__id__in=target_user_ids,
        ).values_list("user_id", flat=True)
        self._queryset = self.get_queryset().exclude(id__in=exclude_user_ids)

    def get_count_from_queryset(self, message_class) -> int:
        """
        Given a Message Class instance (by default drip.drips.DripMessage),
        returns the amount of sent Drips.
        """
        # TODO: try to reduce the side-effects of this method.
        count = 0
        for user in self.get_queryset():
            message_instance = message_class(self, user)
            try:
                result = message_instance.message.send()
                if result:
                    SentDrip.objects.create(
                        drip=self.drip_model,
                        user=user,
                        from_email=self.from_email,
                        from_email_name=self.from_email_name,
                        subject=message_instance.subject,
                        body=message_instance.body,
                    )
                    count += 1
            except Exception as e:
                logging.error(
                    "Failed to send drip {drip} to user {user}: {err}".format(
                        drip=self.drip_model.id,
                        user=str(user),
                        err=str(e),
                    )
                )
        return count

    def send(self) -> int:
        """
        Send the message to each user on the queryset.

        Create SentDrip for each user that gets a message.

        Returns count of created SentDrips.
        """

        if not self.from_email:
            self.from_email = getattr(
                settings,
                "DRIP_FROM_EMAIL",
                settings.DEFAULT_FROM_EMAIL,
            )
        MessageClass = message_class_for(self.drip_model.message_class)

        return self.get_count_from_queryset(MessageClass)

    ####################
    #   USER DEFINED   #
    ####################

    def queryset(self) -> BaseManager:
        """
        Returns a queryset of auth.User who meet the
        criteria of the drip.

        Alternatively, you could create Drips on the fly
        using a queryset builder from the admin interface...
        """
        User = get_user_model()
        return User.objects
